/*
 * Copyright (c) 2003, 2004 Jussi Koskela
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
package fi.iki.hsivonen.webcache.front;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;

import org.mortbay.http.HttpFields;
import org.mortbay.http.HttpResponse;

public class CachedFile {
    
    /** Content type */
    private String contentType;

    /** Content length */
    private volatile int contentLength;

    /** Life time of the content */
    private long lifeTime;
    
    /** Last modified*/
    private long lastModified;
    
    /** Response code from the content server */
    private int responseCode;

    /** Redirect location */
    private String redirectLocation;

    /** Handle to the file */
    private File file;

    /** Number of active readers */
    private int readerCount;

    /** File downloader thread */
    private DownloaderThread downloaderThread;

    /** Flag indicating if this file is invalidated */
    private boolean invalidated;

    /** Back reference to cache handler */
    private CacheHandler cacheHandler;

    /** Url of this file */
    private String url;

    /** Boolean flag indicating if this file is already downloader */
    private boolean downloaded;

    /** Boolean flag indicating if the file is being downloaded */
    private boolean downloading;

    /** Boolean flag indicating if the download failed */
    private boolean downloadFailed;

    /** Boolean flag indicating if the file information is fetched */
    private boolean fetched;

    /** Number of bytes requested for this file */
    private volatile int diskUsage;

    /** Number of bytes currently downloaded from the content server */
    private int bytesReady;

    /** Synchronization variables */
    private Object fetchLock;
    private Object diskWriteLock;
    private Object accessLock;

    /** Dependencies for this file */
    private LinkedHashSet dependencySet;

    /** Invalidation keys that were received during the fetch request */
    private LinkedHashSet invalidationSet;

    /** Custom HTTP header field for dependency keys */
    private static final 
        String DEP_KEYS_FIELD = "X-Cache-Dependencies";
    /** Custom HTTP header field for expiration time */
    private static final 
        String DEP_EXPIRES_FIELD = "X-Cache-Expires";

    /**
     * Class constructor.
     * @param ch the cache handler
     * @param url the URL of this file
     */
    public CachedFile(CacheHandler ch, String url) {
        cacheHandler = ch;
        this.url = url;
        invalidated = false;
        fetched = false;
        downloaded = false;
        downloadFailed = false;
        responseCode = -1;
        redirectLocation = "";
        file = null;
        downloaderThread = null;
        downloading = false;
        diskUsage = 0;
        bytesReady = 0;
        readerCount = 0;
        fetchLock = new Object();
        diskWriteLock = new Object();
        accessLock = new Object();

        dependencySet = new LinkedHashSet();
        invalidationSet = new LinkedHashSet();
    }


    /**
     * Initializes this file. Initialization is done outside the class
     * constructor in order to prevent IO-operations blocking another
     * threads.
     */
    public void delayedInit() throws IOException {
        synchronized (fetchLock) {
            if (!fetched) {
                fetch();
            }
        }
    }
    
    /**
     * Fetches the file information from the content server and launches
     * the file content downloader thread.
     */
    private void fetch() throws IOException {
        FrontServer.debug("Fetch " + url);

        HttpURLConnection connection = 
            (HttpURLConnection) new URL(url).openConnection();
                
        connection.setRequestMethod("GET");
        connection.setUseCaches(false);
        connection.setDoOutput(false);
        connection.setDoInput(true);
        connection.setInstanceFollowRedirects(false);
        
        cacheHandler.recordInvalidations(this);
        
        connection.connect();
        
        contentType = connection.getContentType();
        contentLength = connection.getContentLength();
        lastModified = connection.getLastModified();
        lifeTime = connection.getExpiration() - connection.getDate();
        if (lifeTime <= 0) {
            // mark the file as already expired
            lifeTime = -1;
        }
        
        // read the invalidation keys
        Map m = connection.getHeaderFields();
        List l = (List)m.get(DEP_KEYS_FIELD);
        FrontServer.debug("Keys " + l);
        if (l != null) {
            LinkedHashSet ds = new LinkedHashSet();
            Iterator iter = l.iterator();
            while (iter.hasNext()) {
                String key = (String)iter.next();
                ds.add(key);
            }
            dependencySet = ds;
        }

        // register the global dependencies
        cacheHandler.registerDependencies(this, dependencySet);
        
        // read the expiration date
        long expires = connection.getHeaderFieldDate(DEP_EXPIRES_FIELD,-1);
        if (expires != -1) {
            cacheHandler.registerExpiration(this, new Date(expires));
        }

        // check that this file does not contain invalidated keys
        Iterator iter = invalidationSet.iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            if (dependencySet.contains(key)) {
                cacheHandler.invalidateURL(url);
                break;
            }
        }
        
        responseCode = connection.getResponseCode();
        FrontServer.debug("Response code " + responseCode);
        // save the redirect location if appropriate
        if ((responseCode == HttpResponse.__301_Moved_Permanently) ||
            (responseCode == HttpResponse.__302_Moved_Temporarily)) {
            redirectLocation = 
                connection.getHeaderField(HttpFields.__Location);
        }
        
        // RFC 2616, 4.3
        if ((responseCode - responseCode % 100 != 100) &&
            (responseCode != 204) &&
            (responseCode != 304) &&
            ((contentLength != -1) || (contentType != null))) {
            
            file = File.createTempFile("cms", null, 
                                       cacheHandler.getDirectory());
            file.deleteOnExit();
            
            InputStream in;
            try {
                in = connection.getInputStream();
            } catch (IOException ioe) {
                in = connection.getErrorStream();
            }
            
            if (in != null) {
                downloaderThread = new DownloaderThread(this, in);
                downloading = true;
                downloaderThread.start();
            } else {
                onFinishDownloading(false);
            }
        } else {
            if (responseCode != 302) {
                cacheHandler.invalidateURL(url);
            }
        }
        
        fetched = true;
    }

    /**
     * Registers an invalidation key. This method is invoked by the
     * cache handler only during the fetch opearation.
     * @param key the invalidation key
     */
    public void addInvalidation(String key) {
        synchronized (invalidationSet) {
            invalidationSet.add(key);
        }
    }

    /**
     * Returns the URL of this file.
     * @return the URL
     */
    public String getURL() {
        return url;
    }

    /**
     * Deletes this file from the disk.
     */
    private void delete() {
        if (file != null) {
            FrontServer.debug("File deleted " + url);
            file.delete();
            file = null;
            cacheHandler.releaseDiskSpace(diskUsage);
        }
    }


    /**
     * Returns the content type.
     * @return the content type
     */
    public String getContentType() {
        return contentType;
    }

    /**
     * Returns the content length.
     * @return the content length
     */
    public int getContentLength() {
        return contentLength;
    }

    /**
     * Returns the life time.
     * @return the life time
     */
    public long getLifeTime() {
        return lifeTime;
    }

    /**
     * Returns the last modified value.
     * @return the last modified value
     */
    public long getLastModified() {
        return lastModified;
    }

    /**
     * Returns the content server response code.
     * @return the content server response code
     */
    public int getResponseCode() {
        return responseCode;
    }

    /**
     * Returns the redirect location.
     * @return the redirect location
     */
    public String getRedirectLocation() {
        return redirectLocation;
    }

    /**
     * Invalidates the file.
     */
    public void invalidate() {
        synchronized (accessLock) {
            FrontServer.debug("Invalidating " + url);
            invalidated = true;
            cacheHandler.removeDependencies(this, dependencySet);
            cacheHandler.cancelExpiration(this);
            
            if (readerCount == 0) {
                if (downloading) {
                    downloaderThread.kill();
                } else {
                    delete();
                }
            }
        }
    }
    
    /**
     * Acquires read access to this file.
     */
    public void acquire() {
        synchronized (accessLock) {
            readerCount++; 
            FrontServer.debug("Acquire " + url + " readers " + readerCount);
        }
    }

    /**
     * Releases read access to this file.
     */
    public void release() {
        synchronized (accessLock) {
            readerCount--;
            FrontServer.debug("Release " + url + " readers " + readerCount);
            if (invalidated && !downloading && (readerCount == 0)) {
                delete();
            }
        }
    }

    /**
     * Returns the file output stream.
     * @return the file output stream
     */
    public OutputStream getOutputStream() throws FileNotFoundException {
        return new FileOutputStream(file);
    }

    /**
     * Returns the file input stream.
     * @return the file input stream
     */
    public InputStream getInputStream() throws FileNotFoundException {
        if (file == null) FrontServer.debug("file already invalidated");
        return new FileInputStream(file);
    }

    /**
     * Requests disk space for this file.
     * @param bytes the number of bytes requested
     * @return boolean value indicating the success of the operation
     */
    public boolean requestDiskSpace(int bytes) {
        cacheHandler.requestDiskSpace(bytes);
        diskUsage += bytes;
        return true;
    }

    /**
     * Returns the currently allocated disk space for this file.
     * @return the number of bytes allocated for the file
     */
    public int getDiskUsage() {
        return diskUsage;
    }

    /**
     * Blocks until there is some more content available.
     * @param pos the position from which the data will be read
     * @return boolean value indicating the success of the operation
     */
    public boolean waitInput(int pos) {
        synchronized (diskWriteLock) {
            if (downloadFailed) {
                FrontServer.debug("Download failed");
                return false;
            }

            if (downloaded && (pos == bytesReady)) {
                FrontServer.debug("All data sent");
                return false;
            }

            if (pos < bytesReady) {
                FrontServer.debug("More data already available");
                return true;
            } else {
                while (pos >= bytesReady) {
                    if (downloaded && (pos == bytesReady)) {
                        return true;
                    } else {
                        FrontServer.debug("Wait for more data");
                        try {
                            diskWriteLock.wait();
                        } catch (InterruptedException ie) {
                            System.err.println(ie);
                        }
                    }
                }
                return true;
            }
        }
    }

    /**
     * Notifies other threads that there is more content available in
     * the file input stream.
     * @param length the length of the data written to the disk
     */
    public void notifyDiskWrite(int length) {
        synchronized (diskWriteLock) {
            bytesReady += length;
            diskWriteLock.notifyAll();
        }
    }

    /**
     * Updates the state of the file after the downloader thread has finished
     * its task.
     * @param stopped boolean value indicating if the downloading was
     * cancelled
     */
    public void onFinishDownloading(boolean stopped) {
        synchronized (diskWriteLock) {
            synchronized (accessLock) {
                FrontServer.debug("Download finished " + url + " " + stopped);

                downloading = false;

                downloaded = !stopped;
                downloadFailed = stopped;
                downloaderThread = null;

                if (stopped) {
                    cacheHandler.invalidateURL(url);
                }
                if (invalidated && (readerCount == 0)) {
                    delete();
                } else if (contentLength == -1) {
                    contentLength = bytesReady;
                } else if (contentLength != bytesReady) {
                    downloadFailed = true;
                    cacheHandler.invalidateURL(url);
                }


            }
            diskWriteLock.notifyAll();
        }    
    }
}
