/*
 * Copyright (c) 2003, 2004 Jussi Koskela
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
package fi.iki.hsivonen.webcache.front;
 
import java.io.FileInputStream;
import java.net.URL;
import java.util.Properties;

public class FrontServerConfiguration {

    /** cache server port */
    int serverPort;

    /** cache server invalidator port */
    int invalidatorPort;

    /** minimum number of server threads */
    int minThreads;

    /** maximum number of server threads */
    int maxThreads;

    /** cache directory */
    String cacheDirectory;

    /** maximum allowed disk usage in bytes */
    long maxDiskUsage;

    /** maximum allowed number of cached files */
    int maxCacheFiles;

    /** server tag for the http responses */
    String serverTag;

    /** logging flag */
    boolean logging;

    /** log files */
    String logFile;

    /** content server host */
    String contentServerHost;

    /** content server port */
    int contentServerPort;

    /** front server properties */
    Properties properties;

    /** error messages */
    String errors;

    /** number of errors */
    int errorCount;

    /**
     * Class constructor
     * @param configurationFile the name of the front server configuration file
     */
    public FrontServerConfiguration(String configurationFile) 
    throws InvalidConfigurationException{
        errors = new String();
        errorCount = 0;

        Properties defaults = new Properties();
        defaults.put("ServerPort", "80");
        defaults.put("InvalidatorPort", "8080");
        defaults.put("MaxThreads", "256");
        defaults.put("MinThreads", "4");
        defaults.put("CacheDirectory", "/tmp/cmscache");
        defaults.put("MaxDiskUsage", "1024*1024*1024");
        defaults.put("MaxCacheFiles", "1024*1024");
        defaults.put("Logging", "true");
        defaults.put("Logfile", "access.log");
        defaults.put("ServerTag", "WebCMS Cache");

        properties = new Properties(defaults);

	try {
	    FileInputStream fis = new FileInputStream(configurationFile);
	    properties.load(fis);
	} catch (Exception e) {
            throw new InvalidConfigurationException("Cannot read the configuration file " + configurationFile);
	}

        serverPort = getIntegerProperty("ServerPort", 0, 65535);
        invalidatorPort = getIntegerProperty("InvalidatorPort", 0, 65535);

        if (serverPort == invalidatorPort) {
            addError("InvalidatorPort must not be equal to ServerPort");
        }

        maxThreads = getIntegerProperty("MaxThreads", 1, 1024);
        minThreads = getIntegerProperty("MinThreads", 1, 1024);

        if (minThreads > maxThreads) {
            addError("MinThreads must be less than or equal to MaxThreads");
        }

        cacheDirectory = getStringProperty("CacheDirectory");

        maxDiskUsage = getLongProperty("MaxDiskUsage", 0, Long.MAX_VALUE);
        maxCacheFiles = getIntegerProperty("MaxCacheFiles", 0, Integer.MAX_VALUE);
        serverTag = getStringProperty("ServerTag");

        logging = getBooleanProperty("Logging");
        logFile = getStringProperty("LogFile");

        contentServerHost = getStringProperty("ContentServerHost");

        if (contentServerHost != null) {
            try {
                URL url = new URL("http", contentServerHost, 80, "");
            } catch (Exception e) {
                addError("ContentServerHost does not have a valid host name as its value.");
            }
        }

        contentServerPort = getIntegerProperty("ContentServerPort", 0, 65535);

        if (errorCount > 0) {
            throw new InvalidConfigurationException("Found " + errorCount + " error(s) in the configuration file " + configurationFile + " :\n" + errors);
        }
    }

    /**
     * Appends a new error message.
     * @param error the error message
     */
    private void addError(String error) {
        errorCount++;
        errors = errors + "\t" + error + "\n";
    }

    /**
     * Returns the integer value of a property.
     * @param key the name of the property
     * @param min the minimum value for the property
     * @param max the maximum value for the property
     */
    private int getIntegerProperty(String key, int min, int max) {
        String value = getStringProperty(key);
        int val = 0;
        if (value != null) {
            try {
                val = Integer.parseInt(value);
                if ((val < min) || (val > max)) {
                    addError(key + " does not have an integer value within range " + min + ".." + max);
                }
            } catch (NumberFormatException nfe) {
                addError(key + " does not have an integer value within range" + min + ".." + max);
            }
        }
        return val;
    }

    /**
     * Returns the long value of a property.
     * @param key the name of the property
     * @param min the minimum value for the property
     * @param max the maximum value for the property
     */
    private long getLongProperty(String key, long min, long max) {
        String value = getStringProperty(key);
        long val = 0;
        if (value != null) {
            try {
                val = Long.parseLong(value);
                if ((val < min) || (val > max)) {
                    addError(key + " does not have an integer value within range " + min + ".." + max);
                }
            } catch (NumberFormatException nfe) {
                addError(key + " does not have an integer value within range " + min + ".." + max);
            }
        }
        return val;
    }

    /**
     * Returns the boolean value of a property.
     * @param key the name of the property
     */
    private boolean getBooleanProperty(String key) {
        String value = getStringProperty(key);
        boolean val = false;
        if (value != null) {
            if (value.equalsIgnoreCase("true")) {
                val = true;
            } else if (value.equalsIgnoreCase("false")) {
                val = false;
            } else {
                addError(key + " must have a boolean value 'True' or 'False'");
            }
        }
        return val;
    }

    /**
     * Returns the string value of a property.
     * @param key the name of the property
     */
    private String getStringProperty(String key) {
        String value = properties.getProperty(key);
        if (value == null) {
            addError(key + " is not specified");
        }
        return value;
    }

    /**
     * Returns the cache server port
     * @return the cache server port
     */
    public int getServerPort() {
        return serverPort;
    }

    /**
     * Returns the cache server invalidator port
     * @return the cache server invalidator port
     */
    public int getInvalidatorPort() {
        return invalidatorPort;
    }

    /**
     * Returns the minimum number of server threads.
     * @return the minimum number of server threads
     */
    public int getMinThreads() {
        return minThreads;
    }

    /**
     * Returns the maximum number of server threads.
     * @return the maximum number of server threads
     */
    public int getMaxThreads() {
        return maxThreads;
    }

    /**
     * Returns the cache directory.
     * @return the cache directory
     */
    public String getCacheDirectory() {
        return cacheDirectory;
    }

    /**
     * Returns the maximum allowed disk usage in bytes.
     * @return the maximum disk usage
     */
    public long getMaxDiskUsage() {
        return maxDiskUsage;
    }

    /**
     * Returns the maximum allowed number of cached files.
     * @return the maximum number of cached files
     */
    public int getMaxCacheFiles() {
        return maxCacheFiles;
    }

    /**
     * Returns the server tag.
     * @return the server tag
     */
    public String getServerTag() {
        return serverTag;
    }

    /**
     * Returns the content server host name.
     * @return the content server host name
     */
    public String getContentServerHost() {
        return contentServerHost;
    }

    /**
     * Returns the content server port.
     * @return the content server port
     */
    public int getContentServerPort() {
        return contentServerPort;
    }

    /**
     * Returns the logging flag.
     * @return the logging flag
     */
    public boolean getLogging() {
        return logging;
    }

    /**
     * Returns the name of the log file.
     * @return the name of the log file
     */
    public String getLogFile() {
        return logFile;
    }
}
